from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import wx

from aio_wx_widgets import type_annotations as T
from aio_wx_widgets.containers import Section
from aio_wx_widgets.containers.grid import VERTICAL, Grid
from aio_wx_widgets.core.binding import Binding
from aio_wx_widgets.core.data_types import HorAlign
from aio_wx_widgets.panels.panel import SimplePanel
from aio_wx_widgets.widgets import LabelledItem
from aio_wx_widgets.widgets.button import AioButton
from aio_wx_widgets.widgets.select import Select
from aio_wx_widgets.widgets.text import Text
from aio_wx_widgets.widgets.text_entry import Entry

if TYPE_CHECKING:
    from demo.controller.controller_two import ControllerTwo

_LOGGER = logging.getLogger(__name__)


class ViewTwo(SimplePanel["ControllerTwo"]):
    """Second view."""

    def __init__(self, parent, controller: ControllerTwo):
        self._controller = controller
        super().__init__(parent, controller)
        self._text = Text("This is text that changes as a result of an event happening")

        # self._controller.add_task(self._controller.add_to_log.subscribe(self._add_to_log))
        self.controller.add_to_log.publish += self._add_to_log

        self._callback_text = Text("Nothing selected yet")
        # self._bound_selected_item = Text("Nothing selected yet")

    def populate(self):
        with self.add(Section("Labelled items.")) as sec:
            sec.add(Text("default labelled item:"))
            sec.add(
                LabelledItem(
                    "Item",
                    Select(
                        self.controller.choices,
                        on_select_callback=self._on_select_callback,
                    ),
                )
            )

            l_item = LabelledItem(
                "Item",
                Select(self._controller.choices, binding=self.bind("selected_item")),
                align_right=False,
            )
            sec.add(Text(str(l_item)))
            sec.add(l_item)
            l_item = LabelledItem(
                "Item",
                Select(self._controller.choices, binding=self.bind("selected_item")),
                item_weight=2,
            )
            sec.add(Text(str(l_item)))
            sec.add(l_item)
            l_item = LabelledItem(
                "Item",
                Select(self._controller.choices, binding=self.bind("selected_item")),
                item_weight=2,
                item_alignment=HorAlign.left,
            )
            sec.add(Text(str(l_item)))
            sec.add(l_item)
        with self.add(Section("Selectable, bound to None")) as sec:
            sec.add(
                Select(self._controller.choices, binding=self.bind("selected_item2"))
            )
            sec.add(Text(binding=self.bind("selected_item2")))

        self.add(
            Entry(binding=Binding(self._controller, "value_1")),
            margin=(10, 10, 5, 20),
        )
        self.add(Entry(binding=Binding(self._controller, "value_1")))
        self.add(
            AioButton("test button with normal function as callback", self._press),
            margin=(10, 10, 5, 20),
        )
        self.add(self._text)

        # It is possible to add bare wx-python widgets.
        self.add(wx.StaticLine(parent=self.ui_item), create=False)

        with self.add(Grid()) as grd:
            grd.add(
                Text("This is a selection dropdown"),
                weight=4,
                hor_align=HorAlign.right,
            )
            with grd.add(Grid(orientation=VERTICAL), weight=6) as grd:
                grd.add(
                    Select(
                        self.controller.choices,
                        on_select_callback=self._on_select_callback,
                        binding=self.bind("selected_item"),
                    ),
                    hor_align=HorAlign.left,
                )
                grd.add(Text("Selection based on callback:"))
                grd.add(self._callback_text)
                grd.add(Text("Selection based on binding value:"))
                grd.add(Text(binding=self.bind("selected_item")))
                # grd.add(
                #     AioButton(
                #         label="Display bound value", callback=self._show_bound_value
                #     )
                # )
                # grd.add(self._bound_selected_item)

    # def _show_bound_value(self, evt):
    #     self._bound_selected_item.set_text(self._controller.selected_item)

    def _on_select_callback(self, choice: T.Choice):
        self._callback_text.set_text(str(choice))

    def _press(self, evt):
        self._add_to_log("button pressed.")

    def _add_to_log(self, item: str):
        self._text.set_text(item)



import logging
from pathlib import Path

from demo.controller.image_view_controller import ImageViewController

from aio_wx_widgets.panels.panel import SimplePanel
from aio_wx_widgets.widgets import Text
from aio_wx_widgets.widgets.experimental.fixed_grid import DynamicGrid, FixedGrid
from aio_wx_widgets.widgets.image import Image

_LOGGER = logging.getLogger(__name__)

IMAGE_FOLDER = (Path(__file__).parent) / "image"


class ImageView(SimplePanel[ImageViewController]):
    """A text wrapping view test."""

    def populate(self):
        # self.add(Image(IMAGE_FOLDER / "phoenix_main.png"), weight=1)
        with self.add(DynamicGrid()) as grd:
            # grd.add_space()
            grd.add(Text("this is text"), col=1)
            grd.add(Image(IMAGE_FOLDER / "phoenix_main.png"), col=2, span=2)
            # grd.add(Text("this is text"), col=2)
        with self.add(FixedGrid()) as grd:
            # grd.add_space()
            grd.add(Text("this is text"))
            grd.add(Image(IMAGE_FOLDER / "phoenix_main.png"))
            grd.add(Text("this is text"))
        with self.add(DynamicGrid()) as grd:
            grd.add(Text("This is other test"), col=0)
            grd.add(Image(IMAGE_FOLDER / "phoenix_main.png"), col=1, span=2)



"""Base widget"""

from __future__ import annotations

import logging
from typing import Generic, TypeVar, Union

from aio_wx_widgets.core.binding import Binding, OneWayBindable, TwoWayBindable

_LOGGER = logging.getLogger(__name__)

WxItem = TypeVar("WxItem")


class CallableItem:
    """An item that is callable. (Able to be added to a container)."""

    def init(self, parent):
        """Initialize this aio-item"""
        raise NotImplementedError()

    @property
    def ui_item(self):
        """A wxPython element."""
        raise NotImplementedError()


class BaseWidget(CallableItem, Generic[WxItem]):
    """All widgets should inherit from this."""

    def __init__(
        self,
        ui_item: WxItem,
        min_width: int,
        value_binding: Union[OneWayBindable, TwoWayBindable, None],
        enabled: Union[bool, Binding] = True,
    ):
        self._enabled = enabled
        self._ui_item = ui_item
        self._value_binding = value_binding
        self._min_width = min_width

        self._enabled_binding = None
        if isinstance(self._enabled, Binding):
            self._enabled_binding = OneWayBindable(
                self._enabled, self._set_enabled_value
            )

    @property
    def ui_item(self) -> WxItem:
        return self._ui_item

    def _set_enabled_value(self, enabled):
        if enabled:
            self.ui_item.Enable()
        else:
            self.ui_item.Disable()

    def _make_bindings(self):
        if self._value_binding:
            self._value_binding.make_binding()
        if self._enabled_binding:
            self._enabled_binding.make_binding()

    def _init(self):
        self._make_bindings()
        if self._min_width > -1:
            self.ui_item.SetMinSize((self._min_width, -1))
            # self.ui_item.SetMaxSize((-1,-1))
        # if self._min_width > -1:
        #
        #     actual_x_value = self.ui_item.GetSize()[0]
        #     if actual_x_value < self._min_width:
        #         self.ui_item.SetMinSize((self._min_width, -1))

    def init(self, parent):
        raise NotImplementedError()

    def __call__(self, parent):
        raise NotImplementedError()



"""Text widgets."""

import logging
from typing import Optional, Union

import wx
from wx.lib.wordwrap import wordwrap

from aio_wx_widgets.colors import GREEN
from aio_wx_widgets.const import is_debugging
from aio_wx_widgets.core.base_widget import BaseWidget
from aio_wx_widgets.core.binding import Binding, OneWayBindable

_LOGGER = logging.getLogger(__name__)

__all__ = ["Text"]


def _get_font_info(current_font: wx.Font, font_size: float = 1, bold=False) -> wx.Font:
    font = wx.Font()

    if font_size != 1:
        current_font_size = current_font.GetPointSize()
        new_font_size = int(font_size * current_font_size)
        _LOGGER.debug(
            "Debug ratio of %s results in absolute font size of %s",
            font_size,
            new_font_size,
        )
        font.SetPointSize(new_font_size)

    if bold:
        font.MakeBold()

    return font


class Text(BaseWidget["wx.StaticText"]):
    """Static text."""

    HOR_CENTER = wx.ALIGN_CENTRE_HORIZONTAL
    HOR_RIGHT = wx.ALIGN_RIGHT
    HOR_LEFT = wx.ALIGN_LEFT

    def __init__(
        self,
        text="",
        font_size: float = 1,
        color: Union[int, Binding, None] = None,
        enabled: Union[bool, Binding] = True,
        binding: Optional[Binding] = None,
        bold=False,
        wrap=False,
        min_width=-1,
        hor_align: Optional[int] = None,
    ):
        """Init.

        Args:
            text:
            font_size: Font size ratio.
            wrap: apply text wrapping.

        "normal" text wrappig is not working properly. AFter calling ui_item.Wrap(),
        the text gets extra line breaks where need. If the item is then sized bigger
        the line breaks persist, making the text area too narrow. To overcome this
        you need to `ui_item.SetLabel` again to set the text and after that call
        `ui_item.Wrap()` again. This makes the parent container resize twice: Once
        with the full width text after `SetLabel` and once more after `Wrap` this makes
        the screen very "nervous" and sometimes doesn't scale well at all.

        The workaround for this is to use an intermediate ClientDC instance which mimics
        the statictext box and returns a wrapped string. This string is then used once
        in the `SetLabel` method.
        """
        value_binding = (
            OneWayBindable(binding, self._set_ui_value) if binding is not None else None
        )

        super().__init__(
            wx.StaticText(),
            min_width=min_width,
            value_binding=value_binding,
            enabled=enabled,
        )
        self._parent = None
        self._color = color
        self._text = text
        self._wrap = wrap

        self._font = _get_font_info(self.ui_item.GetFont(), font_size, bold=bold)
        self._color_binding = None
        self._previous_size = (0, 0)
        self._hor_align = hor_align

    def _set_ui_value(self, value):
        self.set_text(value)

    def _get_style(self) -> int:
        style = 0

        if self._hor_align:
            style = style | self._hor_align

        return style

    def init(self, parent):

        self._parent = parent
        self.ui_item.Create(parent, style=self._get_style())

        if self._font:
            self.ui_item.SetFont(self._font)

        if is_debugging():
            self.ui_item.SetBackgroundColour(GREEN)
        self._init()

        if self._wrap:
            self._parent.Bind(wx.EVT_SIZE, self._on_parent_size)
            self._parent.Bind(wx.EVT_MAXIMIZE, self._on_parent_size)
        else:
            self.set_text(self._text)
        self._parent.PostSizeEvent()

    def __call__(self, parent):
        self.init(parent)
        return self

    def _get_wrapped_text(self, text: str, width):
        client_item = wx.ClientDC(self._parent)
        client_item.Clear()
        client_item.SetFont(self._font)

        txt = wordwrap(str(text), width, client_item)
        return txt

    def _on_parent_size(self, evt):
        evt.Skip()

        size = self.ui_item.Size
        _LOGGER.debug("TExt size %s", size)
        if size[0] == 0:
            _LOGGER.debug("Size is zero. Not setting text.")

        elif self._previous_size[0] == size[0]:
            _LOGGER.debug("New size identical to previous size. skipping")

        else:
            self._previous_size = size
            smaller_size = (size[0], -1)
            self._set_text(client_size=smaller_size)

    def _set_text(self, color=None, client_size=None):
        if client_size and self._wrap:
            txt = self._get_wrapped_text(self._text, client_size[0])
        else:
            txt = str(self._text)
        self.ui_item.SetLabel(txt)
        self.ui_item.GetParent().Layout()

    def set_text(self, text, color=None):
        """Set or change the text."""
        self._text = str(text)
        self._set_text(color)

    def _set_color(self, color):
        """Set text color when color binding changes."""
        _LOGGER.debug("Setting color to : %s", color)

        self.ui_item.SetForegroundColour(color)
        self._parent.Refresh()

    def _make_bindings(self):
        if self._color is not None and isinstance(self._color, Binding):
            self._color_binding = OneWayBindable(self._color, self._set_color)
            self._color_binding.make_binding()
        super()._make_bindings()



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

from aio_wx_widgets.containers.grid import Grid
from aio_wx_widgets.containers.group import Section
from aio_wx_widgets.panels.panel import SimplePanel
from aio_wx_widgets.widgets.text import Text

if TYPE_CHECKING:
    from demo.controller.controller_one import ControllerOne

_LOGGER = logging.getLogger(__name__)

IMAGE_FOLDER = (Path(__file__).parent) / "image"


class ViewOne(SimplePanel["ControllerOne"]):
    def __init__(self, parent, controller: "ControllerOne"):
        super().__init__(parent, controller, scrollable=False)

    def populate(self):
        """Populate this view."""

        with self.add(Section("Min width")) as grp:
            with grp.add(Grid()) as grd:
                grd.add(Text("This is a text exceeding 80 min size", min_width=80))
                grd.add(Text("Other text."))

            with grp.add(Grid()) as grd:
                grd.add(Text("text < 80", min_width=80))
                grd.add(Text("Other text."))



import logging

from aio_wx_widgets.controller import BaseController

_LOGGER = logging.getLogger(__name__)


class TextWrappingController(BaseController):
    def __init__(self, model):
        self.color_red = None
        super().__init__(model)



"""Panel. Sits in a frame. Holds widgets."""
from __future__ import annotations

import logging
from typing import Generic, TypeVar

import wx
from wx.lib.scrolledpanel import ScrolledPanel

from aio_wx_widgets.core.sizers import PanelMixin, SizerMixin

_LOGGER = logging.getLogger(__name__)

#  pylint: disable=invalid-name
C = TypeVar("C")  # noqa


class SimplePanel(PanelMixin, SizerMixin, Generic[C]):
    """A simple panel.

    Consider this the main canvas to which widgets can be attached.
    """

    @property
    def ui_item(self):
        """Return UI item."""
        return self._ui_item

    @property
    def _sizer(self):
        return self._sizer_

    def __init__(self, parent, controller: C, scrollable=False, **kwargs):
        """Init."""
        self._controller = controller
        if scrollable:
            self._ui_item = ScrolledPanel(parent)
            self._ui_item.SetupScrolling()
        else:
            self._ui_item = wx.Panel(parent)
        super().__init__()

        self._sizer_ = wx.BoxSizer(wx.VERTICAL)
        self.ui_item.SetSizer(self._sizer)
        self._ui_item.Parent.Bind(wx.EVT_SIZE, self._on_size)

    @property
    def controller(self) -> C:
        """Return the controller."""
        return self._controller

    def _on_size(self, evt):
        evt.Skip()

        # Ugly as hell. Makes wrapping text better though.
        # to check, open the "text wrapping window" in the demo. Maximize the screen
        # and normalize it again. Observer the difference between textbox sizes when
        # enabling or disabling this code below.
        wx.CallLater(1, self._refresh)
        wx.CallLater(2, self._refresh)
        wx.CallLater(3, self._refresh)

        self._refresh()

    def _refresh(self):
        _LOGGER.debug("posting size event")
        if self._ui_item:
            self._ui_item.PostSizeEvent()



import logging

from demo.controller import text_wrapping_controller
from demo.model.demo_model import DemoModel
from demo.views import text_wrapping

from aio_wx_widgets.frame import DefaultFrame

_LOGGER = logging.getLogger(__name__)


class TextWrappingWindow(DefaultFrame):
    def __init__(self):
        super().__init__("Text wrapping window")

        model = DemoModel()
        controller = text_wrapping_controller.TextWrappingController(model)
        view = text_wrapping.TextWrappingView(self, controller)
        view.populate()
